@@ -45,16 +45,18 @@ from libcloud.storage.types import ContainerIsNotEmptyError
 from libcloud.storage.types import ObjectDoesNotExistError
 from libcloud.storage.types import ObjectHashMismatchError
 from libcloud.storage.types import InvalidContainerNameError
-from libcloud.common.types import LazyList
 from libcloud.common.openstack import OpenStackBaseConnection
 from libcloud.common.openstack import OpenStackDriverMixin
 
-from libcloud.common.rackspace import (
-    AUTH_URL_US, AUTH_URL_UK)
+from libcloud.common.rackspace import AUTH_URL
 
 CDN_HOST = 'cdn.clouddrive.com'
 API_VERSION = 'v1.0'
 
+# Keys which are used to select a correct endpoint from the service catalog.
+INTERNAL_ENDPOINT_KEY = 'internalURL'
+PUBLIC_ENDPOINT_KEY = 'publicURL'
+
 
 class CloudFilesResponse(Response):
     valid_response_codes = [httplib.NOT_FOUND, httplib.CONFLICT]
@@ -97,58 +99,121 @@ class CloudFilesRawResponse(CloudFilesResponse, RawResponse):
     pass
 
 
-class CloudFilesConnection(OpenStackBaseConnection):
+class OpenStackSwiftConnection(OpenStackBaseConnection):
+    """
+    Connection class for the OpenStack Swift endpoint.
+    """
+
+    responseCls = CloudFilesResponse
+    rawResponseCls = CloudFilesRawResponse
+
+    auth_url = AUTH_URL
+    _auth_version = '1.0'
+
+    # TODO: Reverse the relationship - Swift -> CloudFiles
+    def __init__(self, user_id, key, secure=True, **kwargs):
+        # Ignore this for now
+        kwargs.pop('use_internal_url', None)
+        super(OpenStackSwiftConnection, self).__init__(user_id, key,
+                                                       secure=secure,
+                                                       **kwargs)
+        self.api_version = API_VERSION
+        self.accept_format = 'application/json'
+
+        self._service_type = self._ex_force_service_type or 'object-store'
+        self._service_name = self._ex_force_service_name or 'swift'
+
+        if self._ex_force_service_region:
+            self._service_region = self._ex_force_service_region
+        else:
+            self._service_region = None
+
+    def get_endpoint(self, *args, **kwargs):
+        if '2.0' in self._auth_version:
+            endpoint = self.service_catalog.get_endpoint(
+                service_type=self._service_type,
+                name=self._service_name,
+                region=self._service_region)
+        elif ('1.1' in self._auth_version) or ('1.0' in self._auth_version):
+            endpoint = self.service_catalog.get_endpoint(
+                name=self._service_name, region=self._service_region)
+
+        if PUBLIC_ENDPOINT_KEY in endpoint:
+            return endpoint[PUBLIC_ENDPOINT_KEY]
+        else:
+            raise LibcloudError('Could not find specified endpoint')
+
+    def request(self, action, params=None, data='', headers=None, method='GET',
+                raw=False, cdn_request=False):
+        if not headers:
+            headers = {}
+        if not params:
+            params = {}
+
+        self.cdn_request = cdn_request
+        params['format'] = 'json'
+
+        if method in ['POST', 'PUT'] and 'Content-Type' not in headers:
+            headers.update({'Content-Type': 'application/json; charset=UTF-8'})
+
+        return super(OpenStackSwiftConnection, self).request(
+            action=action,
+            params=params, data=data,
+            method=method, headers=headers,
+            raw=raw)
+
+
+class CloudFilesConnection(OpenStackSwiftConnection):
     """
     Base connection class for the Cloudfiles driver.
     """
 
-    auth_url = AUTH_URL_US
     responseCls = CloudFilesResponse
     rawResponseCls = CloudFilesRawResponse
 
-    def __init__(self, user_id, key, secure=True, **kwargs):
+    auth_url = AUTH_URL
+    _auth_version = '2.0'
+
+    def __init__(self, user_id, key, secure=True,
+                 use_internal_url=False, **kwargs):
         super(CloudFilesConnection, self).__init__(user_id, key, secure=secure,
                                                    **kwargs)
         self.api_version = API_VERSION
         self.accept_format = 'application/json'
         self.cdn_request = False
-
-        if self._ex_force_service_region:
-            self.service_region = self._ex_force_service_region
+        self.use_internal_url = use_internal_url
 
     def get_endpoint(self):
-        # First, we parse out both files and cdn endpoints
-        # for each auth version
+        region = self._ex_force_service_region.upper()
+
+        if self.use_internal_url:
+            endpoint_type = 'internal'
+        else:
+            endpoint_type = 'external'
+
         if '2.0' in self._auth_version:
-            eps = self.service_catalog.get_endpoints(
+            ep = self.service_catalog.get_endpoint(
                 service_type='object-store',
-                name='cloudFiles')
-            cdn_eps = self.service_catalog.get_endpoints(
-                service_type='object-store',
-                name='cloudFilesCDN')
-        elif ('1.1' in self._auth_version) or ('1.0' in self._auth_version):
-            eps = self.service_catalog.get_endpoints(name='cloudFiles')
-            cdn_eps = self.service_catalog.get_endpoints(name='cloudFilesCDN')
+                name='cloudFiles',
+                region=region,
+                endpoint_type=endpoint_type)
+            cdn_ep = self.service_catalog.get_endpoint(
+                service_type='rax:object-cdn',
+                name='cloudFilesCDN',
+                region=region,
+                endpoint_type=endpoint_type)
+        else:
+            raise LibcloudError(
+                'Auth version "%s" not supported' % (self._auth_version))
 
         # if this is a CDN request, return the cdn url instead
         if self.cdn_request:
-            eps = cdn_eps
+            ep = cdn_ep
 
-        if self.service_region:
-            _eps = []
-            for ep in eps:
-                if ep['region'].lower() == self.service_region.lower():
-                    _eps.append(ep)
-            eps = _eps
-
-        if len(eps) == 0:
+        if not ep or not ep.url:
             raise LibcloudError('Could not find specified endpoint')
 
-        ep = eps[0]
-        if 'publicURL' in ep:
-            return ep['publicURL']
-        else:
-            raise LibcloudError('Could not find specified endpoint')
+        return ep.url
 
     def request(self, action, params=None, data='', headers=None, method='GET',
                 raw=False, cdn_request=False):
@@ -167,56 +232,12 @@ class CloudFilesConnection(OpenStackBaseConnection):
             action=action,
             params=params, data=data,
             method=method, headers=headers,
-            raw=raw)
-
-
-class CloudFilesUSConnection(CloudFilesConnection):
-    """
-    Connection class for the Cloudfiles US endpoint.
-    """
-
-    auth_url = AUTH_URL_US
-
-
-class CloudFilesUKConnection(CloudFilesConnection):
-    """
-    Connection class for the Cloudfiles UK endpoint.
-    """
-
-    auth_url = AUTH_URL_UK
-
-
-class CloudFilesSwiftConnection(CloudFilesConnection):
-    """
-    Connection class for the Cloudfiles Swift endpoint.
-    """
-
-    def __init__(self, *args, **kwargs):
-        self.region_name = kwargs.pop('ex_region_name', None)
-        super(CloudFilesSwiftConnection, self).__init__(*args, **kwargs)
-
-    def get_endpoint(self, *args, **kwargs):
-        if '2.0' in self._auth_version:
-            endpoint = self.service_catalog.get_endpoint(
-                service_type='object-store',
-                name='swift',
-                region=self.region_name)
-        elif ('1.1' in self._auth_version) or ('1.0' in self._auth_version):
-            endpoint = self.service_catalog.get_endpoint(
-                name='swift', region=self.region_name)
-
-        if 'publicURL' in endpoint:
-            return endpoint['publicURL']
-        else:
-            raise LibcloudError('Could not find specified endpoint')
+            raw=raw, cdn_request=cdn_request)
 
 
 class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
     """
-    Base CloudFiles driver.
-
-    You should never create an instance of this class directly but use US/US
-    class.
+    CloudFiles driver.
     """
     name = 'CloudFiles'
     website = 'http://www.rackspace.com/'
@@ -225,11 +246,26 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
     hash_type = 'md5'
     supports_chunked_encoding = True
 
-    def __init__(self, *args, **kwargs):
-        OpenStackDriverMixin.__init__(self, *args, **kwargs)
-        super(CloudFilesStorageDriver, self).__init__(*args, **kwargs)
+    def __init__(self, key, secret=None, secure=True, host=None, port=None,
+                 region='ord', use_internal_url=False, **kwargs):
+        """
+        @inherits:  :class:`StorageDriver.__init__`
 
-    def list_containers(self):
+        :param region: ID of the region which should be used.
+        :type region: ``str``
+        """
+        # This is here for backard compatibility
+        if 'ex_force_service_region' in kwargs:
+            region = kwargs['ex_force_service_region']
+
+        self.use_internal_url = use_internal_url
+        OpenStackDriverMixin.__init__(self, (), **kwargs)
+        super(CloudFilesStorageDriver, self).__init__(key=key, secret=secret,
+                                                      secure=secure, host=host,
+                                                      port=port, region=region,
+                                                      **kwargs)
+
+    def iterate_containers(self):
         response = self.connection.request('')
 
         if response.status == httplib.NO_CONTENT:
@@ -239,10 +275,6 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
 
         raise LibcloudError('Unexpected status code: %s' % (response.status))
 
-    def list_container_objects(self, container):
-        value_dict = {'container': container}
-        return LazyList(get_more=self._get_more, value_dict=value_dict)
-
     def get_container(self, container_name):
         container_name_encoded = self._encode_container_name(container_name)
         response = self.connection.request('/%s' % (container_name_encoded),
@@ -260,7 +292,7 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
     def get_object(self, container_name, object_name):
         container = self.get_container(container_name)
         container_name_encoded = self._encode_container_name(container_name)
-        object_name_encoded = self._encode_container_name(object_name)
+        object_name_encoded = self._encode_object_name(object_name)
 
         response = self.connection.request('/%s/%s' % (container_name_encoded,
                                                        object_name_encoded),
@@ -275,8 +307,9 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
         raise LibcloudError('Unexpected status code: %s' % (response.status))
 
     def get_container_cdn_url(self, container):
-        container_name = container.name
-        response = self.connection.request('/%s' % (container_name),
+        # pylint: disable=unexpected-keyword-arg
+        container_name_encoded = self._encode_container_name(container.name)
+        response = self.connection.request('/%s' % (container_name_encoded),
                                            method='HEAD',
                                            cdn_request=True)
 
@@ -285,7 +318,7 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
             return cdn_url
         elif response.status == httplib.NOT_FOUND:
             raise ContainerDoesNotExistError(value='',
-                                             container_name=container_name,
+                                             container_name=container.name,
                                              driver=self)
 
         raise LibcloudError('Unexpected status code: %s' % (response.status))
@@ -296,17 +329,18 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
 
     def enable_container_cdn(self, container, ex_ttl=None):
         """
-        @inherits: L{StorageDriver.enable_container_cdn}
+        @inherits: :class:`StorageDriver.enable_container_cdn`
 
-        @param ex_ttl: cache time to live
-        @type ex_ttl: C{int}
+        :param ex_ttl: cache time to live
+        :type ex_ttl: ``int``
         """
-        container_name = container.name
+        container_name = self._encode_container_name(container.name)
         headers = {'X-CDN-Enabled': 'True'}
 
         if ex_ttl:
             headers['X-TTL'] = ex_ttl
 
+        # pylint: disable=unexpected-keyword-arg
         response = self.connection.request('/%s' % (container_name),
                                            method='PUT',
                                            headers=headers,
@@ -427,14 +461,15 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
         """
         Purge edge cache for the specified object.
 
-        @param email: Email where a notification will be sent when the job
+        :param email: Email where a notification will be sent when the job
         completes. (optional)
-        @type email: C{str}
+        :type email: ``str``
         """
         container_name = self._encode_container_name(obj.container.name)
         object_name = self._encode_object_name(obj.name)
         headers = {'X-Purge-Email': email} if email else {}
 
+        # pylint: disable=unexpected-keyword-arg
         response = self.connection.request('/%s/%s' % (container_name,
                                                        object_name),
                                            method='DELETE',
@@ -447,7 +482,7 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
         """
         Get meta data
 
-        @rtype: C{dict}
+        :rtype: ``dict``
         """
         response = self.connection.request('', method='HEAD')
 
@@ -494,18 +529,19 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
         """
         Enable serving a static website.
 
-        @param container: Container instance
-        @type container: L{Container}
+        :param container: Container instance
+        :type container: :class:`Container`
 
-        @param index_file: Name of the object which becomes an index page for
+        :param index_file: Name of the object which becomes an index page for
         every sub-directory in this container.
-        @type index_file: C{str}
+        :type index_file: ``str``
 
-        @rtype: C{bool}
+        :rtype: ``bool``
         """
         container_name = container.name
         headers = {'X-Container-Meta-Web-Index': index_file}
 
+        # pylint: disable=unexpected-keyword-arg
         response = self.connection.request('/%s' % (container_name),
                                            method='POST',
                                            headers=headers,
@@ -518,17 +554,18 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
         Set a custom error page which is displayed if file is not found and
         serving of a static website is enabled.
 
-        @param container: Container instance
-        @type container: L{Container}
+        :param container: Container instance
+        :type container: :class:`Container`
 
-        @param file_name: Name of the object which becomes the error page.
-        @type file_name: C{str}
+        :param file_name: Name of the object which becomes the error page.
+        :type file_name: ``str``
 
-        @rtype: C{bool}
+        :rtype: ``bool``
         """
         container_name = container.name
         headers = {'X-Container-Meta-Web-Error': file_name}
 
+        # pylint: disable=unexpected-keyword-arg
         response = self.connection.request('/%s' % (container_name),
                                            method='POST',
                                            headers=headers,
@@ -541,13 +578,14 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
         Set the metadata header X-Account-Meta-Temp-URL-Key on your Cloud
         Files account.
 
-        @param key: X-Account-Meta-Temp-URL-Key
-        @type key: C{str}
+        :param key: X-Account-Meta-Temp-URL-Key
+        :type key: ``str``
 
-        @rtype: C{bool}
+        :rtype: ``bool``
         """
         headers = {'X-Account-Meta-Temp-URL-Key': key}
 
+        # pylint: disable=unexpected-keyword-arg
         response = self.connection.request('',
                                            method='POST',
                                            headers=headers,
@@ -563,18 +601,19 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
         wish.  This method is specifically for allowing users to retrieve
         or update an object.
 
-        @param obj: The object that you wish to make temporarily public
-        @type obj: L{Object}
+        :param obj: The object that you wish to make temporarily public
+        :type obj: :class:`Object`
 
-        @param method: Which method you would like to allow, 'PUT' or 'GET'
-        @type method: C{str}
+        :param method: Which method you would like to allow, 'PUT' or 'GET'
+        :type method: ``str``
 
-        @param timeout: Time (in seconds) after which you want the TempURL
+        :param timeout: Time (in seconds) after which you want the TempURL
         to expire.
-        @type timeout: C{int}
+        :type timeout: ``int``
 
-        @rtype: C{bool}
+        :rtype: ``bool``
         """
+        # pylint: disable=no-member
         self.connection._populate_hosts_and_request_paths()
         expires = int(time() + timeout)
         path = '%s/%s/%s' % (self.connection.request_path,
@@ -622,6 +661,7 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
         object_name_encoded = self._encode_object_name(object_name)
         request_path = '/%s/%s' % (container_name_encoded, object_name_encoded)
 
+        # pylint: disable=no-member
         headers = {'X-Auth-Token': self.connection.auth_token,
                    'X-Object-Manifest': '%s/%s/' %
                                         (container_name_encoded,
@@ -652,30 +692,63 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
 
         return obj
 
-    def _get_more(self, last_key, value_dict):
-        container = value_dict['container']
+    def list_container_objects(self, container, ex_prefix=None):
+        """
+        Return a list of objects for the given container.
+
+        :param container: Container instance.
+        :type container: :class:`Container`
+
+        :param ex_prefix: Only get objects with names starting with ex_prefix
+        :type ex_prefix: ``str``
+
+        :return: A list of Object instances.
+        :rtype: ``list`` of :class:`Object`
+        """
+        return list(self.iterate_container_objects(container,
+                                                   ex_prefix=ex_prefix))
+
+    def iterate_container_objects(self, container, ex_prefix=None):
+        """
+        Return a generator of objects for the given container.
+
+        :param container: Container instance
+        :type container: :class:`Container`
+
+        :param ex_prefix: Only get objects with names starting with ex_prefix
+        :type ex_prefix: ``str``
+
+        :return: A generator of Object instances.
+        :rtype: ``generator`` of :class:`Object`
+        """
         params = {}
+        if ex_prefix:
+            params['prefix'] = ex_prefix
 
-        if last_key:
-            params['marker'] = last_key
+        while True:
+            container_name_encoded = \
+                self._encode_container_name(container.name)
+            response = self.connection.request('/%s' %
+                                               (container_name_encoded),
+                                               params=params)
 
-        response = self.connection.request('/%s' % (container.name),
-                                           params=params)
+            if response.status == httplib.NO_CONTENT:
+                # Empty or non-existent container
+                break
+            elif response.status == httplib.OK:
+                objects = self._to_object_list(json.loads(response.body),
+                                               container)
 
-        if response.status == httplib.NO_CONTENT:
-            # Empty or inexistent container
-            return [], None, True
-        elif response.status == httplib.OK:
-            objects = self._to_object_list(json.loads(response.body),
-                                           container)
+                if len(objects) == 0:
+                    break
 
-            # TODO: Is this really needed?
-            if len(objects) == 0:
-                return [], None, True
+                for obj in objects:
+                    yield obj
+                params['marker'] = obj.name
 
-            return objects, objects[-1].name, False
-
-        raise LibcloudError('Unexpected status code: %s' % (response.status))
+            else:
+                raise LibcloudError('Unexpected status code: %s' %
+                                    (response.status))
 
     def _put_object(self, container, object_name, upload_func,
                     upload_func_kwargs, extra=None, file_path=None,
@@ -685,6 +758,7 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
         object_name_encoded = self._encode_object_name(object_name)
         content_type = extra.get('content_type', None)
         meta_data = extra.get('meta_data', None)
+        content_disposition = extra.get('content_disposition', None)
 
         headers = {}
         if meta_data:
@@ -692,6 +766,9 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
                 key = 'X-Object-Meta-%s' % (key)
                 headers[key] = value
 
+        if content_disposition is not None:
+            headers['Content-Disposition'] = content_disposition
+
         request_path = '/%s/%s' % (container_name_encoded, object_name_encoded)
         result_dict = self._upload_object(
             object_name=object_name, content_type=content_type,
@@ -751,16 +828,11 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
         return name
 
     def _to_container_list(self, response):
-        # @TODO: Handle more then 10k containers - use "lazy list"?
-        containers = []
-
+        # @TODO: Handle more than 10k containers - use "lazy list"?
         for container in response:
             extra = {'object_count': int(container['count']),
                      'size': int(container['bytes'])}
-            containers.append(Container(name=container['name'], extra=extra,
-                                        driver=self))
-
-        return containers
+            yield Container(name=container['name'], extra=extra, driver=self)
 
     def _to_object_list(self, response, container):
         objects = []
@@ -805,7 +877,10 @@ class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
         return obj
 
     def _ex_connection_class_kwargs(self):
-        return self.openstack_connection_kwargs()
+        kwargs = self.openstack_connection_kwargs()
+        kwargs['ex_force_service_region'] = self.region
+        kwargs['use_internal_url'] = self.use_internal_url
+        return kwargs
 
 
 class CloudFilesUSStorageDriver(CloudFilesStorageDriver):
@@ -815,26 +890,31 @@ class CloudFilesUSStorageDriver(CloudFilesStorageDriver):
 
     type = Provider.CLOUDFILES_US
     name = 'CloudFiles (US)'
-    connectionCls = CloudFilesUSConnection
-
-
-class CloudFilesSwiftStorageDriver(CloudFilesStorageDriver):
-    """
-    Cloudfiles storage driver for the OpenStack Swift.
-    """
-    type = Provider.CLOUDFILES_SWIFT
-    name = 'CloudFiles (SWIFT)'
-    connectionCls = CloudFilesSwiftConnection
 
     def __init__(self, *args, **kwargs):
-        self._ex_region_name = kwargs.get('ex_region_name', 'RegionOne')
-        super(CloudFilesSwiftStorageDriver, self).__init__(*args, **kwargs)
+        kwargs['region'] = 'ord'
+        super(CloudFilesUSStorageDriver, self).__init__(*args, **kwargs)
 
-    def openstack_connection_kwargs(self):
-        rv = super(CloudFilesSwiftStorageDriver,
-                   self).openstack_connection_kwargs()
-        rv['ex_region_name'] = self._ex_region_name
-        return rv
+
+class OpenStackSwiftStorageDriver(CloudFilesStorageDriver):
+    """
+    Storage driver for the OpenStack Swift.
+    """
+    type = Provider.CLOUDFILES_SWIFT
+    name = 'OpenStack Swift'
+    connectionCls = OpenStackSwiftConnection
+
+    # TODO: Reverse the relationship - Swift -> CloudFiles
+
+    def __init__(self, key, secret=None, secure=True, host=None, port=None,
+                 region=None, **kwargs):
+        super(OpenStackSwiftStorageDriver, self).__init__(key=key,
+                                                          secret=secret,
+                                                          secure=secure,
+                                                          host=host,
+                                                          port=port,
+                                                          region=region,
+                                                          **kwargs)
 
 
 class CloudFilesUKStorageDriver(CloudFilesStorageDriver):
@@ -844,7 +924,10 @@ class CloudFilesUKStorageDriver(CloudFilesStorageDriver):
 
     type = Provider.CLOUDFILES_UK
     name = 'CloudFiles (UK)'
-    connectionCls = CloudFilesUKConnection
+
+    def __init__(self, *args, **kwargs):
+        kwargs['region'] = 'lon'
+        super(CloudFilesUKStorageDriver, self).__init__(*args, **kwargs)
 
 
 class FileChunkReader(object):
@@ -896,7 +979,7 @@ class ChunkStreamReader(object):
             raise StopIteration
 
         block_size = self.chunk_size
-        if self.bytes_read + block_size >\
+        if self.bytes_read + block_size > \
                 self.end_block - self.start_block:
             block_size = self.end_block - self.start_block - self.bytes_read
             self.stop_iteration = True
